package src.BAALL

import akka.actor.ActorDSL._
import akka.actor.ActorLogging
import scala._
import scala.collection.parallel.mutable.ParHashMap
import src.ship._

/**
 * Created with IntelliJ IDEA.
 * User: dodi
 * Date: 05.09.12
 * Time: 14:08
 * propagates updates in the ontology via the lknx protocol
 * lknxobjects together with their current values. the hashmap is used to compute the diffs
 * it is currently not supported to add objects at runtime, even though this should be possible in principle
 */

class LKNXUpdater(server: String, port: Int, OntUpdater: OntUpdater,
                  registeredObjects: ParHashMap[String, LKNXObject], lknx: Linknx)
  extends ActWithStash with ActorLogging {
  val valueMap = Map(
    "DimOn" -> "255",
    "DimOff" -> "0",
    "DimBright" -> "190",
    "DimDark" -> "63",
    "DimHalf" -> "127",
    "On" -> "on", "Off" -> "off",
    "Open" -> "on", "Closed" -> "off")

  log.info("Start LKNXUpdater...")

  override def receive = {
    case UCHPolling(on) => if (on) lknx.writeObject("synchronize", "on") else lknx.writeObject("synchronize", "off")
    case v@EmptyOntology => OntUpdater.ontology = v
    case CurrentOntology(ont: Ontology) =>

      OntUpdater.ontology = ont.copy(ont.tbox, ont.rbox, ont.abox)
      log.info("Query3 "+ont.getDiff._2+"\nresult1 = "+ont.query("livingLight1:On")+
        "\nresult2 = "+OntUpdater.ontology.query("livingLight1:On"))
      ont.getDiff match {
        case (dels, adds) => {
          log.info("get an ontology update and send information to yamamoto\n" + registeredObjects)
          for (add <- adds) {
            add match {
              case ConceptAssertion(Const(name), ConceptConstant(cname)) =>
                // println("process Concept " + add + " " + name + " cname:" + cname)
                registeredObjects.get(name) match {
                  case Some(obj) => if (!obj.isInstanceOf[PosObject]) {
                    val newvalue = valueMap.get(cname).getOrElse(cname)
                    if (lknx.getVal(obj) != newvalue) {
                      //TEST registeredObjects.put(name, lknx.modifyObject(obj,newvalue))
                      lknx.writeObject(name, newvalue)
                    }
                  }
                  case _ => //println("object " + name + " does not exist in BAALL")
                }
              case RoleAssertion(Const(pos), Const(name), NamedRole("position_status")) =>
                if (!ont.query(ConceptConstant("WheelChair").mkAssertion(Const(name))).isEmpty) {
                  lknx.writeObject(name, pos)

                  ont.query(List(
                    NamedRole("wheelchair_carries").roleAssert(Const(name), Var("person")),
                    ConceptConstant("Person").mkAssertion(Var("person"))
                  )) match {
                    case s :: _ => val pname = s(Var("person")).asInstanceOf[Const].name
                      lknx.writeObject(pname, pos)
                    case _ =>
                  }
                }
              case RoleAssertion(Const(name), Const(pos), NamedRole("at")) =>
                //println("process role assertion (" + name + ", " + pos + "):at")
                registeredObjects.get(name) match {
                  case Some(obj) =>
                    val newvalue = pos

                    if (!ont.query(ConceptAssertion(Const(pos), ConceptConstant("Position"))).isEmpty && lknx.getVal(obj) != newvalue) {
                      // TEST registeredObjects.put(name, lknx.modifyObject(obj,newvalue))
                      lknx.writeObject(name, newvalue)
                      log.info("done write")
                    } else if (lknx.getVal(obj) != newvalue) {
                      log.warning("Name " + pos + " does not denote a valid position on BAALL")
                    }
                  case _ => //println("object " + name + " does not exist in BAALL")
                }
              case x => //println("cannot yet? process " + x + " in update")
            }
          }
          //
          lknx.sendWriteQueue()
        }
      }
    case _ => // println("LKNXUpdate got "+v)
  }
}
