package src.ship

import de.derivo.sparqldlapi.{QueryEngine, QueryBinding, QueryResult, Query}
import org.semanticweb.owlapi.apibinding.OWLManager
import org.semanticweb.owlapi.model.{OWLClass, OWLOntologyManager, PrefixManager, OWLNamedIndividual}
import org.semanticweb.owlapi.reasoner.{Node, OWLReasoner}
import scala.collection.{JavaConversions, mutable}

/**
 * Created by IntelliJ IDEA.
 * User: dodi
 * Date: 7/23/12
 * Time: 10:32 AM
 */

object OWLQuery extends JavaLogging with ConceptParser {
  val factory = OWLManager.createOWLOntologyManager().getOWLDataFactory

  /**
   * returns all individuals of type concept
   * @param concept: the class
   * @param pm: prefixmanager that stores the ontology prefix
   * @param reasoner: owl reasoner to be used to execute the query
   */

  def getIndividuals(concept: Concept, pm: PrefixManager, reasoner: OWLReasoner): List[String] = {

    val temp = reasoner.getInstances(Ontology2OWL.concept2OWLClassExpr(concept, pm, factory, usenegabbrev = true), false)
    //println("temp = "+temp)
    val iter = temp.getNodes.iterator()
    var individuals = List[String]()
    while (iter.hasNext) {
      val node: Node[OWLNamedIndividual] = iter.next()
      //println(node.getRepresentativeElement.getIRI.getFragment)
      individuals = node.getRepresentativeElement.getIRI.getFragment :: individuals
    }
    // log.info("Individuals of " + concept + " = " + individuals)
    individuals
  }

  def isEquivalentConcept(c: ConceptConstant, cexpr: Concept, pm: PrefixManager, reasoner: OWLReasoner): Boolean = {
    val OWLcexpr = Ontology2OWL.concept2OWLClassExpr(cexpr, pm, factory, usenegabbrev = true)
    val OWLc = Ontology2OWL.concept2OWLClassExpr(c, pm, factory, usenegabbrev = true).asInstanceOf[OWLClass]
    //println("Equivalent classes "+cexpr+": "+reasoner.getEquivalentClasses(OWLcexpr))
    reasoner.getEquivalentClasses(OWLcexpr).contains(OWLc)
  }

  def getSuperClasses(c: Concept, pm: PrefixManager, reasoner: OWLReasoner): Set[Concept] = {
    val OWLc = Ontology2OWL.concept2OWLClassExpr(c, pm, factory, usenegabbrev = true).asInstanceOf[OWLClass]
    log.info("Superclasses of " + c + " " + reasoner.getSuperClasses(OWLc, true))
    JavaConversions.asScalaSet(reasoner.getSuperClasses(OWLc, true).getNodes).toSet[Node[OWLClass]] map {
      owlc => ConceptConstant(owlc.getRepresentativeElement.getIRI.getFragment)
    }
  }

  def isLowestLevelClass(c: ConceptConstant, pm: PrefixManager, reasoner: OWLReasoner): Boolean = {
    val OWLc = Ontology2OWL.concept2OWLClassExpr(c, pm, factory, usenegabbrev = true).asInstanceOf[OWLClass]
    log.info("Subclasses of " + c + " " + reasoner.getSubClasses(OWLc, true))
    reasoner.getSubClasses(OWLc, false).isBottomSingleton
  }

  def getRoleMembers(role: Role, pm: PrefixManager, reasoner: OWLReasoner): Set[(String, String)] = {
    val OWLobjectproperty = Ontology2OWL.role2OWLPropExpr(role, pm, factory)
    val domiter: java.util.Iterator[Node[OWLClass]] = reasoner.getObjectPropertyDomains(OWLobjectproperty, true).getNodes.iterator()
    var members = mutable.Set.empty[(String, String)]
    while (domiter.hasNext) {
      val individualsiter = reasoner.getInstances(domiter.next().getRepresentativeElement, true).getNodes.iterator
      while (individualsiter.hasNext) {
        val ind: OWLNamedIndividual = individualsiter.next.getRepresentativeElement
        val coind: java.util.Iterator[Node[OWLNamedIndividual]] = reasoner.getObjectPropertyValues(ind, OWLobjectproperty).getNodes.iterator()
        while (coind.hasNext) {
          members.add(
            (ind.getIRI.getFragment,
              coind.next().getRepresentativeElement.getIRI.getFragment))
        }
      }
    }
    members.toSet
  }


  def getDifferentIndividuals(ind: String, pm: PrefixManager, reasoner: OWLReasoner): Set[String] = {
    /* val domiter : java.util.Iterator[Node[OWLNamedIndividual]] =
      reasoner.getDifferentIndividuals(factory.getOWLNamedIndividual(":"+ind,pm)).getNodes.iterator()
    val members = mutable.Set.empty[String]
    while (domiter.hasNext) {
      members.add(domiter.next().getRepresentativeElement.getIRI.getFragment)
    }
    members.toSet */
    getAllIndividuals(reasoner) -- getSameIndividuals(ind, pm, reasoner)
  }

  def getAllIndividuals(reasoner: OWLReasoner): Set[String] = {
    /* val result : mutable.Set[String] = mutable.Set.empty[String]
    val it = reasoner.getKB.getIndividuals.iterator()
    while (it.hasNext) {
      var i = it.next().getName
      i = i.substring(i.indexOf("#")+1)
      result.add(i)
    }
    //println("all individuals "+result)
    result.toSet[String]  */

    val domiter: java.util.Iterator[OWLNamedIndividual] = reasoner.getInstances(factory.getOWLThing, false).getFlattened.iterator()
    val members = mutable.Set.empty[String]
    while (domiter.hasNext) {
      members.add(domiter.next().getIRI.getFragment)
    }
    members.toSet
  }

  def getSameIndividuals(ind: String, pm: PrefixManager, reasoner: OWLReasoner): Set[String] = {
    val domiter: java.util.Iterator[OWLNamedIndividual] = reasoner.getSameIndividuals(factory.getOWLNamedIndividual(":" + ind, pm)).getEntities.iterator()
    var members = mutable.Set.empty[String]
    while (domiter.hasNext) {
      members.add(domiter.next().getIRI.getFragment)
    }
    members.toSet
  }


  def printSep[A](xs: List[A], sep: String, printFun: A => String): String = {
    xs match {
      case List() => ""
      case x :: y :: ys =>
        printFun(x) + sep + printSep(y :: ys, sep, printFun)
      case x :: List() => printFun(x)
    }
  }

  // aBoxFormula2RDFQueryString(aBoxFormula: ABoxFormula): String unused in Revision 1787

  def aBoxFormula2QueryString(pm: PrefixManager)(aBoxFormula: ABoxFormula): String = {
    val ns: String = pm.getDefaultPrefix
    assert(ns.charAt(ns.length() - 1).equals('#'))
    def i2s(vc: VarOrConstant): String = vc match {
      case Var(s) => "?" + s
      case Const(s) => "<" + ns + s + ">"
      case Literal(s) => if (isLiteral(s)) s else "\"" + s + "\""
    }
    aBoxFormula match {
      case ConceptAssertion(Var(x), concept) =>
        "Type(?" + x + ", " + Ontology2OWL.concept2SparqlString(concept, pm, factory) + ")"
      case ConceptAssertion(Const(x), concept) =>
        "Type(<" + ns + x + ">, " + Ontology2OWL.concept2SparqlString(concept, pm, factory) + ")"
      case RoleAssertion(src, trg, role@DataRole(rn)) =>
        "PropertyValue(" + i2s(src) + ", " +
          Ontology2OWL.datarole2OWLPropExpr(role, pm, factory) + ", " + i2s(trg) + ")"
      case RoleAssertion(src, trg, role) =>
        "PropertyValue(" + i2s(src) + ", " +
          Ontology2OWL.role2OWLPropExpr(role, pm, factory) + ", " + i2s(trg) + ")"
      case DifferentFromAssertion(vc1, vc2) =>
        "DifferentFrom(" + i2s(vc1) + ", " + i2s(vc2) + ")"
      case SameAsAssertion(vc1, vc2) =>
        "SameAs(" + i2s(vc1) + ", " + i2s(vc2) + ")"
      case _ => throw new Exception("aBoxFormula2QueryString " + aBoxFormula)
    }
  }

  def createDisjointnessQuery(pm: PrefixManager, c1: Concept, c2: Concept): de.derivo.sparqldlapi.Query = {
    val ns = pm.getDefaultPrefix
    val concept1 = Ontology2OWL.concept2SparqlString(c1, pm, factory)
    val concept2 = Ontology2OWL.concept2SparqlString(c2, pm, factory)
    //"DisjointWith("+concept1+", "+concept2+")"

    val queryString = "PREFIX ship: <" + ns + ">\n" +
      "SELECT * \n" +
      " WHERE { DisjointWith(" + concept1 + ", " + concept2 + ") }"
    //println("ns is "+ns)
    //println("formulas are :"+aBoxFormulas)
    //println(queryString)
    de.derivo.sparqldlapi.Query.create(queryString)

  }

  /**
   * takes a list of preconditions (and) and returns all possible substitutions
   *
   */
  def aBoxFormulas2Query(aBoxFormulas: List[ABoxFormula], pm: PrefixManager): de.derivo.sparqldlapi.Query = {
    val ns = pm.getDefaultPrefix
    val vars = aBoxFormulas flatMap (a => a.freeVars)
    val queryString =
      if (vars.length == 0) "PREFIX ship: <" + ns + ">\n" + "ASK {" + printSep(aBoxFormulas map aBoxFormula2QueryString(pm), ",\n ", (s: String) => s) + " }"
      else {
        "SELECT " + (if (vars.distinct.isEmpty) "*" else printSep(vars.distinct, " ", (s: ObjectLevelVariable) => "?" + s.toString)) +
          " WHERE {\n" +
          printSep(aBoxFormulas map aBoxFormula2QueryString(pm), ",\n ", (s: String) => s) + " }"
      }
    //println("ns is "+ns)
    //println("formulas are :"+aBoxFormulas)
    //println(queryString)
    de.derivo.sparqldlapi.Query.create(queryString)
  }

  /**
   * returns the super classes of the specified concept
   * @param concept name of the concept whose superclasses shall be computed
   * @param reasoner reasoner to perform the query
   * @param ns namespace, e.g. http://www.semanticweb.org/ontologies/2012/6/SHIP.owl#
   */
  def getDirectSuperClasses(concept: Concept, manager: OWLOntologyManager, reasoner: OWLReasoner, ns: String): List[VarOrConstant] = {

    //println(concept.toString)
    val squery: Query = Query.create("PREFIX ship: <" + ns + ">\n" +
      "SELECT ?c WHERE {\n"
      + " Class(?c), DirectSubClassOf(ship:" + concept.toString + ",?c)}")

    val engine: QueryEngine = QueryEngine.create(manager, reasoner, true)
    val result: QueryResult = engine.execute(squery)
    queryResult2Substitution(result, squery) map ((subst: Substitution) => subst.apply(Var("c")))
  }

  def getType(individual: String, manager: OWLOntologyManager, reasoner: OWLReasoner, ns: String): List[VarOrConstant] = {
    //println("gettype of "+individual)
    val squery: Query = Query.create("PREFIX ship: <" + ns + ">\n" +
      "SELECT ?c WHERE {\n"
      + " Type(ship:" + individual + ",?c)}")

    val engine: QueryEngine = QueryEngine.create(manager, reasoner, true)
    val result: QueryResult = engine.execute(squery)
    queryResult2Substitution(result, squery) map ((subst: Substitution) => subst.apply(Var("c")))
  }

  def execQuery(query: String, manager: OWLOntologyManager, reasoner: OWLReasoner, prefixname: String, ns: String): List[Substitution] = {
    // println("Execute query...")
    // println(query)
    val squery: Query = Query.create("PREFIX " + prefixname + ": <" + ns + ">\n" + query)
    val engine: QueryEngine = QueryEngine.create(manager, reasoner, true)
    val result: QueryResult = engine.execute(squery)
    queryResult2Substitution(result, squery)
  }


  def queryResult2Substitution(result: QueryResult, query: Query): List[Substitution] = {
    var res = mutable.MutableList[Substitution]()

    if (query.isAsk) {
      if (result.ask()) {
        val subst: Substitution = Id()
        res += subst
        //println("Query is positive ask ")
      } //else //println("Query is negative ask ")
    }
    else {
      var iter = result.iterator()
      while (iter.hasNext) {
        var subst: Substitution = Id()
        val qbinding: QueryBinding = iter.next()
        if (qbinding.size() > 0) {
          val substiter = qbinding.getBoundArgs.iterator()
          while (substiter.hasNext) {
            val entry = substiter.next()
            val indname = qbinding.get(entry).getValue
            //@warning namespaces are cut off
            subst = CompSubst(/(Const(indname.substring(indname.indexOf("#") + 1)), Var(entry.getValue)), subst)
          }
          res += subst
        }
      }
    }
    res.toList
  }
}
