package src.ship

import akka.actor.ActorDSL._
import akka.actor.{ActorLogging, ReceiveTimeout, ActorRef}
import java.io.File
import scala.Some
import scala.concurrent.duration._
import src.BAALL.{DrivingDestinations, Position}

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 18.06.2012
 * Time: 17:03
 */

abstract class SHIPEngineRoomCommand

case class AddDeclaration(n: String, d: DLDeclaration) extends SHIPEngineRoomCommand

case class AddObserver(a: ActorRef) extends SHIPEngineRoomCommand

case class RemoveObserver(a: ActorRef) extends SHIPEngineRoomCommand

case class StartProgram(p: DLProgram) extends SHIPEngineRoomCommand

case class StartMonitor(phi: FOLTL[Concept, ABoxFormula], desc: String) extends SHIPEngineRoomCommand

case class UpdateAbox(update: List[ABoxFormula], deleteIndividuals: List[Const]) extends SHIPEngineRoomCommand

case class SaveOntology(filename: String)

case class QueryOnt(update: List[ABoxFormula]) extends SHIPEngineRoomCommand

case class Exit() extends SHIPEngineRoomCommand

object Trace extends SHIPEngineRoomCommand


abstract class SHIPEngineRoomMessages {
  def toStringLong = toString
}

case class ProgramUpdateAbox(p: DLProgram, res: DLProgStepResult) extends SHIPEngineRoomMessages

case class ProgramError(p: DLProgram, msg: String) extends SHIPEngineRoomMessages

case class ProgramSuccess(p: DLProgram) extends SHIPEngineRoomMessages

case class ProgramFailure(p: DLProgram) extends SHIPEngineRoomMessages

case class ProgramStutter(p: DLProgram, rest: DLProgram) extends SHIPEngineRoomMessages

case class MonitorFailure(phi: FOLTL[Concept, ABoxFormula],
                          desc: String,
                          lastvalid: FOLTL[Concept, ABoxFormula],
                          currentstatephi: FOLTL[Concept, ABoxFormula],
                          currentstatephiresult: FOLTL[Concept, ABoxFormula]) extends SHIPEngineRoomMessages

case class MonitorProgress(phi: FOLTL[Concept, ABoxFormula], current: FOLTL[Concept, ABoxFormula], desc: String) extends SHIPEngineRoomMessages

case class MonitorSuccess(phi: FOLTL[Concept, ABoxFormula], current: FOLTL[Concept, ABoxFormula], desc: String) extends SHIPEngineRoomMessages

case class MonitorStopAndStartProgram(phi: FOLTL[Concept, ABoxFormula], prog: DLProgram, desc: String) extends SHIPEngineRoomMessages

case class DeclarationAdded(s: String) extends SHIPEngineRoomMessages

case class CurrentOntology(o: Ontology) extends SHIPEngineRoomMessages

case class CurrentTrace(tr: SHIPTrace) extends SHIPEngineRoomMessages {
  override def toString = "CurrentTrace(" + (if (tr.isEmpty) "" else "last: " + tr.last + ",") + "#steps = " + tr.length + ")"

  override def toStringLong = "CurrentTrace:\n" + tr
}

case class ProcessesReset(pl: List[String]) extends SHIPEngineRoomMessages

case class MonitorsReset(ml: List[String]) extends SHIPEngineRoomMessages

case class Log(s: String) extends SHIPEngineRoomMessages

abstract class RollandClientMessage extends SHIPEngineRoomMessages

case class RollandPosition2D(name: String, position: Position) extends RollandClientMessage

case class RollandNewRouteGraph(name: String, graph: DrivingDestinations) extends RollandClientMessage


class SHIPEngineRoom(dls: Map[String, DLDeclaration]) extends ActWithStash with ActorLogging {
  var programqueue = List.empty[Pair[DLProgram, List[DLProgStepResult]]]
  var monitors = Map.empty[FOLTL[Concept, ABoxFormula], List[MonitorExec]]
  var decls = dls
  var ontology: Ontology = Ontology(List.empty[TBoxFormula], List.empty[RBoxFormula], List.empty[ABoxFormula])
  var listeners = Set[ActorRef]()
  val trace = new SHIPTrace(500)

  def inform(msg: SHIPEngineRoomMessages) {
    for (listener <- listeners) {
      listener ! msg
    }
  }

  def progress_all_current_monitors(ont: Ontology, pq: List[Pair[DLProgram, List[DLProgStepResult]]]): List[Pair[DLProgram, List[DLProgStepResult]]] = {
    pq map (x => progress_current_monitors(ont, x))
  }

  def progress_current_monitors(ont: Ontology, entry: Pair[DLProgram, List[DLProgStepResult]]): Pair[DLProgram, List[DLProgStepResult]] = {
    val program = entry._1
    val history = entry._2
    history match {
      case Nil => (program, Nil)
      case progstepresult :: hs =>
        progstepresult match {
          case Stutter(o, step, DLEvaluation(dcls, cont)) =>
            val res = progress_program_monitors(ont, dcls, cont)
            if (res == cont) {
              (program, Stutter(ont, step, DLEvaluation(dcls, cont)) :: history)
            } else {
              (program, Step(ont, step, DLEvaluation(dcls, res)) :: history)
            }
          case Step(o, step, DLEvaluation(dcls, cont)) =>
            val res = progress_program_monitors(ont, dcls, cont)
            if (res == cont) {
              (program, Stutter(ont, step, DLEvaluation(dcls, cont)) :: history)
            } else {
              (program, Step(ont, step, DLEvaluation(dcls, res)) :: history)
            }
          case _ => (program, history)
        }
    }
  }

  def progress_program_monitors(o: Ontology, decls: Map[String, DLDeclaration], p: DLProgram): DLProgram = {
    p match {
      case v@MonitorExec(orig, desc, phi) => phi match {
        case True => v
        case Not(True) => v
        case _ =>
          val d1 = phi match {
            case v@DefPred(Pred(n, actpars)) => DLEvaluation(decls, p).getDescription(n, actpars)
            case _ => desc
          }
          FOLTL.progress(o, phi, decls) match {
            case res@True =>
              inform(MonitorSuccess(orig, res, d1))
              val newmon = MonitorExec(orig, d1, res)
              newmon.lastphi = Some(phi)
              newmon
            case res@Not(True) =>
              val diagnosephi1 = FOLTL.diagnose_formula(ontology, phi)
              inform(MonitorFailure(orig, d1, res, diagnosephi1, FOLTL.progress(o, diagnosephi1, decls)))
              val newmon = MonitorExec(orig, d1, res)
              newmon.lastphi = Some(phi)
              newmon
            case res =>
              //println("a1")
              inform(MonitorProgress(orig, res, d1))
              val newmon = MonitorExec(orig, d1, res)
              newmon.lastphi = Some(phi)
              newmon
          }
      }
      case Seq(p1, p2) => Seq(progress_program_monitors(o, decls, p1), p2)
      case Par(p1, p2) => Par(progress_program_monitors(o, decls, p1), progress_program_monitors(o, decls, p2))
      case Alt(p1, p2) => Alt(progress_program_monitors(o, decls, p1), p2)
      case Star(p1) => Star(progress_program_monitors(o, decls, p1))
      case _ => p
    }
  }

  def eval_some_progstep(pq: List[Pair[DLProgram, List[DLProgStepResult]]],
                         prefix: List[Pair[DLProgram, List[DLProgStepResult]]]): Pair[List[Pair[DLProgram, List[DLProgStepResult]]], Boolean] = {
    log.info("Starting eval_some_progstep")
    pq match {
      case Nil => (prefix, false)
      case (dlprog, history) :: rs =>
        //println("History = "+history)
        val res: Option[DLProgStepResult] = history match {
          case Nil =>
            val evalprogmonitors = progress_program_monitors(ontology, decls, dlprog)
            Some(DLEvaluation(decls, evalprogmonitors).eval_step(ontology))
          case Stutter(o, step, lastdleval) :: _ => // TODO: Check if Ontology has changed meanwhile?
            Some(DLEvaluation(decls, lastdleval.p).eval_step(ontology))
          case Step(o, step, lastdleval) :: _ => Some(DLEvaluation(decls, lastdleval.p).eval_step(ontology))
          case _ => None
        }
        //println("res = "+res)
        res match {
          case Some(p) =>
            for (mon <- p.getStutteringMonitors) inform(MonitorProgress(mon.orig, mon.phi, mon.desc))
            p match {
              case v@Stutter(o, step, dleval) =>
                step match {
                  case x@MonitorExec(orig, desc, phi) => inform(MonitorProgress(orig, phi, desc))
                  case s => s
                }
                inform(ProgramStutter(dlprog, dleval.p))
                eval_some_progstep(rs, prefix ++ List((dlprog, v :: history)))
              case v@Step(o, step, dleval) =>
                step match {
                  case MonitorExec(orig, desc, phi) =>
                    if (phi == True || phi == Not(True)) {
                      trace.enqueue(Pair(step, o))
                      //inform(MonitorSuccess(orig,True()))
                    } // else inform(MonitorProgress(orig,phi,desc))
                  case _ => trace.enqueue(Pair(step, o))
                    if (!o.diffIsEmpty) {
                      inform(ProgramUpdateAbox(dlprog, v))
                      inform(CurrentOntology(o))
                    }
                }
                ontology = o
                (prefix ++ rs ++ List((dlprog, v :: history)), true)
              case v@Error(msg) =>
                inform(ProgramError(dlprog, msg))
                (prefix ++ ((dlprog, v :: history) :: rs), true)
              case v@Fail(o, step) =>
                step match {
                  case MonitorExec(orig, desc, phi) =>
                    val diagnosephi1 = FOLTL.diagnose_formula(ontology, phi)
                    inform(MonitorFailure(orig, desc, phi, diagnosephi1, FOLTL.progress(o, diagnosephi1, decls)))
                    trace.enqueue(Pair(step, o))
                  case s => s
                }
                inform(ProgramFailure(dlprog))
                (prefix ++ rs ++ List((dlprog, v :: history)), true)
              case v@Done(o, step) =>
                ontology = o
                trace.enqueue(Pair(step, o))
                step match {
                  case MonitorExec(orig, desc, phi) => inform(MonitorSuccess(orig, phi, desc))
                  case _ => inform(ProgramUpdateAbox(dlprog, v)); inform(CurrentOntology(o))
                }
                inform(ProgramSuccess(dlprog))
                (prefix ++ rs ++ List((dlprog, v :: history)), true)
            }
          case None => eval_some_progstep(rs, prefix ++ List((dlprog, history)))
        }
    }
  }

  context.setReceiveTimeout(20.millisecond)

  override def receive = {
    case ReceiveTimeout => checkOntology()
    case x: RollandClientMessage => inform(x)
    case Trace => inform(CurrentTrace(trace))
    case Show(n) =>
      if (n == "#ONTOLOGY") inform(Log("Current ontology\n" + ontology.toStringLong))
      else decls.get(n) match {
        case Some(d) => d match {
          case a@Action(_, fpars, pre, exec, eff, conds) =>
            val execstr = exec match {
              case Some(Pair(m, args)) =>
                m.getDeclaringClass.getSimpleName ++ "." ++ m.getName ++
                  (if (args.isEmpty) "" else "(" + args + ")")
              case None => ""
            }
            inform(Log("action " + a + " = {" +
              "\n pre = " + pre +
              "\n effect = " + eff +
              "\n exec = " + execstr +
              "\n " + conds + "\n}"))
          case f@Function(fn, fpars, body) =>
            inform(Log("function " + fn + "(" + fpars + ") = {" +
              body + "}"))
          case _ => inform(Log("declaration " + n + " " + d))
        }
        case None => inform(Log("unknown declaration " + n))
      }
    case StartProgram(p) =>
      programqueue = (p, List[DLProgStepResult]()) :: programqueue
      setChanged()
    case v@StartMonitor(phi, desc) =>
      //println(v)
      val progress: FOLTL[Concept, ABoxFormula] = FOLTL.progress(ontology, phi, decls)
      progress match {
        case Not(True) => {
          val diagnosephi1 = FOLTL.diagnose_formula(ontology, phi)
          inform(MonitorFailure(phi, desc, progress, diagnosephi1, diagnosephi1)) //(FOLTL.progress(ontology,diagnosephi1,decls))))
        }
        case True => inform(MonitorSuccess(phi, progress, desc))
        case RunProg(prog) => {
          monitors = monitors - phi
          inform(MonitorStopAndStartProgram(phi, prog, desc))
          SHIPTool.er ! StartProgram(prog)
        }
        case _ => inform(MonitorProgress(phi, progress, desc))
      }
      monitors = monitors + (phi -> List(MonitorExec(phi, desc, progress)))
      setUnchanged()
    case SaveOntology(filename) =>
      val file = new File(filename)
      ontology.saveToOWLFile(file)
      inform(Log("saved ontology under " + file))
    case QueryOnt(formulas) =>
      val individuals = ontology.allIndividuals
      val knownSubst: List[Substitution] = (formulas flatMap (phi =>
        phi.freeVars)).distinct flatMap (o => o match {
        case v@Var(n) => individuals flatMap (s =>
          if (!v.isExplicitVar && s.equalsIgnoreCase(n)) List(/(Const(s), v)) else Nil)
      })
      val groundingsubst = (knownSubst :\ (Id(): Substitution))((s1, s2) => CompSubst(s1, s2))
      inform(Log("Instantiating by " + groundingsubst))
      val newformulas = formulas map groundingsubst.apply
      val substs: List[Substitution] = ontology.query(newformulas)
      substs match {
        case Nil => inform(Log("Query unsatisfiable " + newformulas))
        case _ => for (subst <- substs) inform(Log("Query satisfiable " + newformulas + "; subst = " + subst))
      }
    case UpdateAbox(update, deleteIndividuals) =>
      if (!update.isEmpty || !deleteIndividuals.isEmpty) {
        //ontology.interruptReasoner()
        ontology.updatesimple(update, deleteIndividuals) match {
          case Some(o) =>
            ontology = o
            trace.enqueue(Pair(Appl("SHIPUpdate", Nil), ontology))
            if (ontology.isBooleanAbox) {
              inform(CurrentOntology(ontology.toClassicAbox))
            }
            monitors = check_and_update_monitors()
            setChanged()
          case None =>
        }
      }
      inform(CurrentOntology(ontology))
    case AddObserver(a) =>
      //println(AddObserver(a))
      a ! CurrentOntology(ontology)
      listeners = listeners + a
      setUnchanged()
    case RemoveObserver(a) =>
      listeners = listeners - a
      setUnchanged()
    case Exit() =>
      context.stop(self)
      setUnchanged()
    case AddDeclaration(n, d) =>
      decls = decls updated(n, d)
      ontology.addIndirectEffectRules(Map(n -> d))
      setUnchanged()
      inform(DeclarationAdded(n))
    case ResetOntology(o) =>
      o.addIndirectEffectRules(ontology.indirecteffectrules)
      ontology = o
      inform(CurrentOntology(ontology))
      monitors = check_and_update_monitors()
      setChanged()
    case ResetProcesses(pl) =>
      programqueue = if (pl.isEmpty) Nil
      else programqueue filterNot (p => pl contains p._1)
      inform(ProcessesReset(pl))
      setUnchanged()
    case ResetMonitors(ml) =>
      monitors = if (ml.isEmpty) Map.empty[FOLTL[Concept, ABoxFormula], List[MonitorExec]]
      else monitors filterNot (p => ml contains p._1)
      inform(MonitorsReset(ml))
      setUnchanged()
  }

  var changed = false

  def setChanged() {
    changed = true
  }

  def setUnchanged() {
    changed = false
  }

  def setChangedTo(v: Boolean) {
    changed = v
  }

  def isChanged: Boolean = changed

  def checkOntology() {
    if (isChanged) {
      if (ontology.isConsistent) {
        if (!ontology.diffIsEmpty) programqueue = progress_all_current_monitors(ontology, programqueue)
        //println("Evaluating some "+programqueue)
        eval_some_progstep(programqueue, Nil) match {
          case (pq, c) =>
            programqueue = pq
            //println("Setting programmqueue to "+(pq map (p => "["+p._1+" : "+(if (p._2.isEmpty) "No started" else p._2.head))))
            //println("Setting changed to "+c)
            setChangedTo(c)
          case _ => setUnchanged()
        }
      } else {
        inform(Log("Ontology is inconsistent; stopping"))
        setUnchanged()
      }
    }
    //println("Checking input")
    if (isChanged) monitors = check_and_update_monitors()
  }

  def check_and_update_monitors(): Map[FOLTL[Concept, ABoxFormula], scala.List[MonitorExec]] = {
    (monitors.keySet :\ Map.empty[FOLTL[Concept, ABoxFormula], List[MonitorExec]])((phi, map) => {
      val hist = (monitors get phi).get
      hist.head.phi match {
        case Not(True) => map updated(phi, hist)
        case _ => map updated(phi, {
          //println("check_and_update_monitors")
          val res = FOLTL.progress(ontology, hist.head.phi, decls)
          var result = res
          res match {
            case Not(True) =>
              val diagnosephi1 = FOLTL.diagnose_formula(ontology, hist.head.phi)
              inform(MonitorFailure(phi, hist.head.desc, hist.head.phi, diagnosephi1,
                FOLTL.progress(ontology, diagnosephi1, decls)))
            case True => inform(MonitorSuccess(phi, res, hist.head.desc))
            case RunProg(prog) =>
              monitors = monitors - phi
              inform(MonitorStopAndStartProgram(phi, prog, hist.head.desc))
              SHIPTool.er ! StartProgram(prog)
              result = Not(True)
            case _ => inform(MonitorProgress(phi, res, hist.head.desc))
          }
          MonitorExec(hist.head.orig, hist.head.desc, result)
        } :: hist)
      }
    })
  }
}
