package src.ship

import collection.mutable
import scala.Predef.Pair

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 3.09.2012
 * Time: 10:15
 */

class SHIPTrace(maxsize: Int) extends mutable.Queue[Pair[DLProgram, Ontology]] {
  def enqueue(e: Pair[DLProgram, Ontology]) {
    if (size == maxsize) dequeue()
    super.enqueue(e)
  }

  override def toString(): String = {
    val printDiff = (p: Pair[List[ABoxFormula], List[ABoxFormula]]) => {
      val del: List[ABoxFormula] = p._1
      val add: List[ABoxFormula] = p._2
      (del :\ "")((a, b) => "-" + a + ", " + b) +
        (add :\ "")((a, b) => "+" + a + ", " + b)
    }


    val sb = StringBuilder.newBuilder
    val it = this.iterator
    var counter = 1
    while (it.hasNext) {
      val pair = it.next()

      val dlprog = pair._1
      val ont: Ontology = pair._2
      //println(dlprog)
      dlprog match {
        case Act(n, actpars) =>
          sb.append(counter.toString + ". " + dlprog.toString + "> " + printDiff(ont.getDiff) + "\n")
          counter = counter + 1
        case Appl("SHIPUpdate", _) =>
          sb.append(counter.toString + ". INPUT" + "> " + printDiff(ont.getDiff) + "\n")
          counter = counter + 1
        case MonitorExec(orig, desc, True) =>
          val description = if (desc.isEmpty) orig.toString
          else "'" + desc + "' (" + orig + ")"
          sb.append(counter.toString + ". Monitor success " + description + "\n")
          counter = counter + 1
        case MonitorExec(orig, desc, Not(True)) =>
          val description = if (desc.isEmpty) orig.toString
          else "'" + desc + "' (" + orig + ")"
          sb.append(counter.toString + ". Monitor failure " + description + "\n")
          counter = counter + 1
        case v => v
      }
    }
    //println("SB "+sb+" / "+sb.toString())
    sb.toString()
  }
}
