package src.test

import org.junit.runner.RunWith
import org.scalatest.FunSuite
import org.scalatest.junit.JUnitRunner
import org.semanticweb.owlapi.apibinding.OWLManager
import org.semanticweb.owlapi.model.OWLObjectProperty
import src.ship._

@RunWith(classOf[JUnitRunner])
class Ontology2OWLTest extends FunSuite {

  test("Translation of a role declaration") {
    val rboxformula = RoleDeclaration(NamedRole("namedRole"), ConceptConstant("Src"), ConceptConstant("Trg"))
    val pm = Builtin.prefixManager
    val ont = new Ontology(List(), List(rboxformula), List())
    val owlont = Ontology2OWL.convert(ont, pm,
      OWLManager.createOWLOntologyManager())
    val role: OWLObjectProperty = owlont.getOWLOntologyManager.getOWLDataFactory.getOWLObjectProperty("namedRole", pm)
    assert(role.getDomains(owlont).size() > 0)
    assert(role.getRanges(owlont).size() > 0)
  }

  test("Translation of a transitive role") {
    val rboxformula = RoleTra(NamedRole("transRole"))
    val pm = Builtin.prefixManager
    val ont = new Ontology(List(), List(rboxformula), List())
    val owlont = Ontology2OWL.convert(ont, pm,
      OWLManager.createOWLOntologyManager())
    val role: OWLObjectProperty = owlont.getOWLOntologyManager.getOWLDataFactory.getOWLObjectProperty("transRole", pm)
    assert(role.isTransitive(owlont))
  }

  test("Translation of a reflexive role") {
    val rboxformula = RoleRef(NamedRole("reflRole"))
    val pm = Builtin.prefixManager
    val ont = new Ontology(List(), List(rboxformula), List())
    val owlont = Ontology2OWL.convert(ont, pm,
      OWLManager.createOWLOntologyManager())
    val role: OWLObjectProperty = owlont.getOWLOntologyManager.getOWLDataFactory.getOWLObjectProperty("reflRole", pm)
    assert(role.isReflexive(owlont))
  }

  test("Translation of a symmetric role") {
    val rboxformula = RoleSym(NamedRole("symRole"))
    val pm = Builtin.prefixManager
    val ont = new Ontology(List(), List(rboxformula), List())
    val owlont = Ontology2OWL.convert(ont, pm,
      OWLManager.createOWLOntologyManager())
    val role: OWLObjectProperty = owlont.getOWLOntologyManager.getOWLDataFactory.getOWLObjectProperty("symRole", pm)
    assert(role.isSymmetric(owlont))
  }

  test("Translation of a irreflexive role") {
    val rboxformula = RoleIrr(NamedRole("irrRole"))
    val pm = Builtin.prefixManager
    val ont = new Ontology(List(), List(rboxformula), List())
    val owlont = Ontology2OWL.convert(ont, pm,
      OWLManager.createOWLOntologyManager())
    val role: OWLObjectProperty = owlont.getOWLOntologyManager.getOWLDataFactory.getOWLObjectProperty("irrRole", pm)
    assert(role.isIrreflexive(owlont))
  }

  test("Translation of a functional role") {
    val rboxformula = RoleFun(NamedRole("funRole"))
    val pm = Builtin.prefixManager
    val ont = new Ontology(List(), List(rboxformula), List())
    val owlont = Ontology2OWL.convert(ont, pm,
      OWLManager.createOWLOntologyManager())
    val role: OWLObjectProperty = owlont.getOWLOntologyManager.getOWLDataFactory.getOWLObjectProperty("funRole", pm)
    assert(role.isFunctional(owlont))
  }

  test("Translation of a inverse functional role") {
    val rboxformula = RoleInvFun(NamedRole("invfunRole"))
    val pm = Builtin.prefixManager
    val ont = new Ontology(List(), List(rboxformula), List())
    val owlont = Ontology2OWL.convert(ont, pm,
      OWLManager.createOWLOntologyManager())
    val role: OWLObjectProperty = owlont.getOWLOntologyManager.getOWLDataFactory.getOWLObjectProperty("invfunRole", pm)
    assert(role.isInverseFunctional(owlont))
  }

  test("Translation of a asymmetric role") {
    val rboxformula = RoleAsym(NamedRole("asymRole"))
    val pm = Builtin.prefixManager
    val ont = new Ontology(List(), List(rboxformula), List())
    val owlont = Ontology2OWL.convert(ont, pm,
      OWLManager.createOWLOntologyManager())
    val role: OWLObjectProperty = owlont.getOWLOntologyManager.getOWLDataFactory.getOWLObjectProperty("asymRole", pm)
    assert(role.isAsymmetric(owlont))
  }

  test("Translation of a subrole") {
    val rboxformula = RoleDeclaration(NamedRole("subRole"), ConceptConstant("Src"), ConceptConstant("Trg"))
    val rboxformula2 = RoleDeclaration(NamedRole("supRole"), ConceptConstant("Src"), ConceptConstant("Trg"))
    val rboxformula3 = RoleInclusion(List(NamedRole("subRole")), NamedRole("supRole"))
    val pm = Builtin.prefixManager
    val ont = new Ontology(List(), List(rboxformula, rboxformula2, rboxformula3), List())
    val owlont = Ontology2OWL.convert(ont, pm,
      OWLManager.createOWLOntologyManager())
    val role: OWLObjectProperty = owlont.getOWLOntologyManager.getOWLDataFactory.getOWLObjectProperty("subRole", pm)
    assert(role.getSuperProperties(owlont).size() > 0)
  }
}
